"""This module does blah blah."""

import random

class BernoulliArm():
    """This class does blah blah."""
    def __init__(self, p):
        self.p = p

    def draw(self):
        """This function does blah blah."""
        if random.random() > self.p:
            return 0.0
        return 1.0



def MonteCarloSimulation(algo, arms, reward_amounts, n_iter, horizon):
    """
    Monte Carlo simulation used to test bandit algorithms with synthetic data

    algo: bandit algorithm of choice
    arms: array of arms to draw from
    n_iter: number of iterations
    horizon: horizon of time to simulate for each iteration
    """
    
    chosen_arms = [0.0 for i in range(n_iter * horizon)]
    rewards = [0.0 for i in range(n_iter * horizon)]
    cumulative_rewards = [0.0 for i in range(n_iter * horizon)]
    sim_nums = [0.0 for i in range(n_iter * horizon)]
    times = [0.0 for i in range(n_iter * horizon)]
    n_arms = len(arms)

    for sim in range(n_iter):
        sim += 1
        algo.initialize(n_arms)
    
        for t in range(horizon):
            t += 1
            index = (sim - 1) * horizon + t - 1
            sim_nums[index] = sim
            times[index] = t
            
            # choose an arm for time t
            chosen_arm = algo.select_arm()
            chosen_arms[index] = chosen_arm
            
            # record reward from chosen arm at time t
            reward = arms[chosen_arms[index]].draw() * reward_amounts[chosen_arm]
            rewards[index] = reward

            # record cumulative rewards
            cumulative_rewards[index] = cumulative_rewards[index - 1] + reward
            
            # update estimated reward from each arm
            algo.update(chosen_arm, reward)
  
    return [sim_nums, times, chosen_arms, rewards, cumulative_rewards]


from ._epsilon_greedy import EpsilonGreedy, ind_max, hello_world
from ._softmax import Softmax, categorical_draw

__all__ = [
    'EpsilonGreedy',
    'Softmax',
    'ind_max',
    'hello_world',
    'categorical_draw'
]


import random

def categorical_draw(probs):
    """
    Given an array if probabilities, returns  
    """
    probs = [0.2, 0.4, 0.1, 0.3]
    z = random.random()
    cum_prob = 0.0
    for i in range(len(probs)):
        prob = probs[i]
        cum_prob += prob
        if cum_prob > z:
            return i
    return len(probs) - 1
    cum_prob

class Softmax():
    """
    Softmax will have a variable exploration rate based on the value specified
    for tau or a "temperature" parameter. Based on the analogy from physics in
    which a system will behave differently at higher temperatures.

    Parameters
    ----------
    tau : float
        Scaling factor that determines how often the Softmax algorithm explores.
        tau = 0 results in choosing the arm with the highest estimated value 
        every time. tau = Inf would result in purely random exploration, similar
        to the Epsilon Greedy algorithm.
    
    counts: int

    values: float

    Attributes
    ----------

    
    """
    def __init__(self, tau, counts, values):
        self.tau = tau
        self.counts = counts
        self.values = values
        return
    
    def initialize(self, n_arms):
        self.counts = [0 for col in range(n_arms)]
        self.values = [0.0 for col in range(n_arms)]
        return
    
    def select_arm(self):
        z = sum([math.exp(v / self.tau) for v in self.values])
        probs = [math.exp(v / self.tau) / z for v in self.values]
        return categorical_draw(probs)

    def update(self, chosen_arm, reward):
        self.counts[chosen_arm] = self.counts[chosen_arm] + 1
        n = self.counts[chosen_arm]
        
        value = self.values[chosen_arm]
        new_value = ((n - 1) / float(n)) * value + (1 / float(n)) * reward
        self.values[chosen_arm] = new_value
        return


import random

def ind_max(x):
    """
    returns the index that corresponds to the maximum value in array x
    """
    m = max(x)
    return x.index(m)

def hello_world():
    """
    test function to see if package is installed in editable mode
    """
    print("hello universe")

class EpsilonGreedy():
    """

    Parameters
    ----------
    epsilon: float
        Percentage of time the bandit explores
    n_arms: int
        Number of arms
    rewards: array
        Average units of reward observed for each successful arm pull
    conv_rates: array
        Success rates for each arm
    counts: array
        number of times each arm has been pulled

    Attributes
    ----------
    values: array
        average number of successes observed for each arm (i.e. conversion rate)
    """
    def __init__(self, epsilon, n_arms, rewards, conv_rates=None, counts=None):
        self.epsilon    = epsilon
        self.rewards    = rewards
        self.conv_rates = [0 for i in range(n_arms)] if conv_rates is None else conv_rates
        self.counts     = [0 for i in range(n_arms)] if counts is None else counts
        self.values     = [i*j for i,j in zip(self.conv_rates,self.rewards)]
        # raise error if n_arms does not equal number of entries in counts or values
        if ((n_arms != len(self.counts)) or (n_arms != len(self.values)) or (n_arms != len(self.conv_rates))):
            raise ValueError("n_arms does not match the length of counts/values/conv_rates")
        return

    def select_arm(self):
        if random.random() > self.epsilon:
            chosen_arm = ind_max(self.values)
        else:
            chosen_arm = random.randrange(len(self.values))
        return chosen_arm

    def update(self, chosen_arm, success_flag):
        # increments counts for chosen arm
        self.counts[chosen_arm] = self.counts[chosen_arm] + 1
        # calculate new average reward for chosen arm
        n = self.counts[chosen_arm]
        prev_value = self.values[chosen_arm]
        new_value = ((n - 1) / float(n)) * prev_value + (1 / float(n)) * success_flag * self.rewards[chosen_arm]
        self.values[chosen_arm] = new_value
        return
    
    def batch_update(self, chosen_arm, num_times_chosen, num_successes, observed_reward=None):
        
        # save previous algo parameter values
        prev_counts     = self.counts.copy()
        prev_conv_rates = self.conv_rates.copy()
        prev_rewards    = self.rewards.copy()
        prev_values     = self.values.copy()

        # increments counts for chosen arm
        self.counts[chosen_arm] = prev_counts[chosen_arm] + num_times_chosen

        # update conversion rates
        self.conv_rates[chosen_arm] = ((prev_conv_rates[chosen_arm] * prev_counts[chosen_arm]) + num_successes) / self.counts[chosen_arm]

        # calculate new average reward for chosen arm
        observed_reward = prev_rewards[chosen_arm] if observed_reward is None else observed_reward
        prev_total_rewards = (prev_rewards[chosen_arm] * prev_conv_rates[chosen_arm] * prev_counts[chosen_arm])
        new_total_rewards  = num_successes * observed_reward
        self.rewards[chosen_arm] = (prev_total_rewards + new_total_rewards) / self.counts[chosen_arm]
        
        # calculate new average value for chosen arm
        self.values[chosen_arm] = (self.conv_rates[chosen_arm] * self.counts[chosen_arm] * self.rewards[chosen_arm])/self.counts[chosen_arm]
        return




# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['banditcoot', 'banditcoot.algorithms', 'banditcoot.arms', 'banditcoot.utility']

package_data = \
{'': ['*']}

install_requires = \
['pandas>=1.5.3,<2.0.0']

setup_kwargs = {
    'name': 'banditcoot',
    'version': '0.1.7',
    'description': 'Python module for developing and testing multi-armed bandits',
    'long_description': 'Banditcoot\n====================\n\n<img align="left" width="75" src="https://github.com/dan-kwon/banditcoot/blob/master/resources/Banditcoot.png"/> \n\n**banditcoot** is a Python module is intended for my own personal use when developing and testing multi-armed bandit algorithms. Forked from the accompanying [code repository](https://github.com/johnmyleswhite/BanditsBook) for [Bandit Algorithms for Website Optimization](https://www.oreilly.com/library/view/bandit-algorithms-for/9781449341565/).\n\n*Any use of this code is at your own risk.*\n</br>\n\n## Installation\n--------------------\nRecommended method of installation is to use pip \n```\npip install banditcoot\n```\n\n\n## Quickstart\n--------------------\n',
    'author': 'dan-kwon',
    'author_email': 'danielkwon02@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
